if __name__ == "__main__":
    print("testing")

    tradingsymbol = "SBIN"
    transaction_type = "BUY"
    exchange = "NSE"
    quantity = 1
    order_type = "LIMIT"
    price = 550
    product = "MIS"
    tag = "test"